"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import itertools as ittl
import typing as h

from babelwidget.main import backend_t, grid_lyt_h
from babelwidget.main import group_h as group_wgt_h
from babelwidget.main import label_h as label_wgt_h
from babelwidget.main import stack_h as stack_wgt_h
from conf_ini_g.interface.storage.section import INI_UNIT_SECTION
from conf_ini_g.interface.window.generic import FormattedName
from conf_ini_g.interface.window.parameter.main import parameter_t, parameters_t
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)
from conf_ini_g.phase.specification.section.controller import controller_t
from conf_ini_g.phase.specification.section.specific import (
    any_section_h as any_section_spec_h,
)
from conf_ini_g.phase.specification.section.specific import (
    controlled_section_t as controlled_section_spec_t,
)
from conf_ini_g.phase.specification.section.specific import (
    free_section_t as free_section_spec_t,
)


@d.dataclass(repr=False, eq=False)
class _section_t:  # Cannot be abstracted.
    HEADER_NAMES: h.ClassVar[tuple[str]] = (
        "Parameter",
        "Type(s)",
        "Value",
        "Unit",
    )
    HEADER_STYLE: h.ClassVar[str] = "background-color: darkgray; padding-left: 5px;"

    spec_name: str
    formatted_name: str
    library_wgt: group_wgt_h

    @classmethod
    def NewWithName(
        cls, name: str, backend: backend_t, /, *, controller: controller_t = None
    ) -> h.Self:
        """"""
        if controller is None:
            controller = ""
        else:
            controller = (
                f" ⮜ {FormattedName(controller.section, ' ')}."
                f"{FormattedName(controller.parameter, ' ')}"
            )
        formatted_name = FormattedName(name, " ") + controller

        output = cls(
            spec_name=name, formatted_name=formatted_name, library_wgt=backend.group_t()
        )
        output.library_wgt.setTitle(formatted_name)

        return output

    @classmethod
    def Headers(cls, backend: backend_t, /) -> h.Sequence[label_wgt_h]:
        """"""
        output = []

        for text in cls.HEADER_NAMES:
            header = backend.label_t(f'<font color="blue">{text}</font>')
            header.setStyleSheet(cls.HEADER_STYLE)
            output.append(header)

        return output

    @property
    def active_parameters(self) -> h.Sequence[parameter_t]:
        """"""
        raise NotImplementedError

    def AsINI(self) -> dict[str, str]:
        """"""
        return {_elm.spec_name: _elm.Text() for _elm in self.active_parameters}


@d.dataclass(repr=False, eq=False)
class free_section_t(_section_t, parameters_t):
    @classmethod
    def NewForSpecification(
        cls,
        section_spec: free_section_spec_t,
        backend: backend_t,
        /,
    ) -> h.Self | None:
        """"""
        output = cls.NewWithName(section_spec.name, backend)

        parameters, _, layout = _SectionParameters(
            section_spec, section_spec.name == INI_UNIT_SECTION, backend
        )
        if parameters.__len__() == 0:
            return None

        output.extend(parameters)

        for h_idx, header in enumerate(cls.Headers(backend)):
            layout.addWidget(header, 0, h_idx)
        output.library_wgt.setLayout(layout)

        return output

    @property
    def all_parameters(self) -> parameters_t:
        """"""
        return self

    @property
    def active_parameters(self) -> parameters_t:
        """"""
        return self


@d.dataclass(repr=False, eq=False)
class controlled_section_t(_section_t, dict[str, parameters_t]):
    controlling_values: list[str] = d.field(init=False, default=None)
    page_stack: stack_wgt_h = d.field(init=False, default=None)

    @classmethod
    def NewForSpecification(
        cls,
        section_spec: controlled_section_spec_t,
        controller: controller_t,
        backend: backend_t,
        /,
    ) -> h.Self | None:
        """"""
        output = cls.NewWithName(section_spec.name, backend, controller=controller)

        controlling_values = []
        page_stack = backend.stack_t()
        for controlling_value, parameter_specs in section_spec.items():
            parameters, _, layout = _SectionParameters(parameter_specs, False, backend)
            if parameters.__len__() == 0:
                continue

            controlling_values.append(controlling_value)
            output[controlling_value] = parameters_t(parameters)

            for h_idx, header in enumerate(cls.Headers(backend)):
                layout.addWidget(header, 0, h_idx)
            page = backend.base_t()
            page.setLayout(layout)
            page_stack.addWidget(page)

        if output.__len__() == 0:
            return None

        output.controlling_values = controlling_values
        output.page_stack = page_stack

        # Curiously, the stacked widget cannot be simply declared as child of instance;
        # This must be specified through a layout.
        layout = backend.hbox_lyt_t()
        layout.addWidget(page_stack)
        layout.setContentsMargins(0, 0, 0, 0)
        output.library_wgt.setLayout(layout)

        return output

    @property
    def all_parameters(self) -> parameters_t:
        """"""
        return parameters_t(ittl.chain(*self.values()))

    @property
    def active_parameters(self) -> parameters_t:
        """"""
        return self[self.controlling_values[self.page_stack.currentIndex()]]


any_section_h = free_section_t | controlled_section_t


def _SectionParameters(
    specifications: any_section_spec_h | h.Iterator[parameter_spec_t],
    section_is_unit: bool,
    backend: backend_t,
    /,
) -> tuple[h.Sequence[parameter_t], h.Sequence[str], grid_lyt_h]:
    """"""
    parameters = []
    parameter_names = []

    layout = backend.grid_lyt_t()
    layout.setAlignment(backend.ALIGNED_TOP)
    layout.setColumnStretch(0, 4)
    layout.setColumnStretch(1, 1)
    layout.setColumnStretch(2, 8)
    layout.setColumnStretch(3, 1)
    layout.setContentsMargins(0, 0, 0, 0)

    for row, parameter_spec in enumerate(specifications, start=1):
        parameter = parameter_t.NewForSpecification(parameter_spec, backend)
        parameters.append(parameter)
        parameter_names.append(parameter_spec.name)

        layout.addWidget(parameter.name, row, 0, alignment=backend.ALIGNED_RIGHT)
        layout.addWidget(parameter.type, row, 1)
        layout.addWidget(parameter.value.library_wgt, row, 2, 1, 2 - 1)
        if not (section_is_unit or (parameter.unit is None)):
            layout.addWidget(parameter.unit, row, 3)

    return parameters, parameter_names, layout


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import textwrap as text
import typing as h

from conf_ini_g.extension.python import SpecificationPath
from conf_ini_g.interface.storage.parameter import INI_UNIT_SEPARATOR
from conf_ini_g.interface.storage.section import INI_UNIT_SECTION
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.parameter.unit import unit_t
from conf_ini_g.phase.specification.parameter.value import MISSING_REQUIRED_VALUE
from conf_ini_g.phase.specification.section.controller import controller_t
from conf_ini_g.phase.specification.section.generic import section_t
from conf_ini_g.phase.specification.section.specific import (
    any_section_h,
    controlled_section_t,
    free_section_t,
    parameters_t,
)
from conf_ini_g.phase.specification.section.unit import IsUnitSection
from logger_36 import LOGGER
from rich.text import Text as text_t
from str_to_obj import ObjectFromStr
from str_to_obj.api.catalog import callable_t, choices_t
from str_to_obj.api.type import ANY_TYPE, type_t
from str_to_obj.interface.console import TypeAsRichStr
from str_to_obj.type.hint import any_hint_h


@d.dataclass(slots=True, repr=False, eq=False)
class config_t(list[any_section_h]):
    sections: d.InitVar[h.Sequence[section_t]]
    path: str = None

    relative_to_home: d.InitVar[bool] = True

    def __post_init__(
        self, sections: h.Sequence[section_t], relative_to_home: bool
    ) -> None:
        """
        Raising exceptions is adapted here since execution cannot proceed without a
        valid specification.
        """
        specific = []
        for section in sections:
            as_dict = dict(
                (_elm.name, getattr(section, _elm.name)) for _elm in d.fields(section)
            )
            if section.controller is None:
                del as_dict["controller"]
                del as_dict["alternatives"]
                section = free_section_t(**as_dict)
            else:
                section = controlled_section_t(**as_dict)
            specific.append(section)

        self.extend(specific)
        self.path = SpecificationPath(sections, relative_to_home=relative_to_home)

        for section in specific:
            if isinstance(section, free_section_t):
                continue

            controller = section.controller
            if self.GetController(controller).type is ANY_TYPE:
                choices = choices_t.NewAnnotatedType(section.controlling_values)
                self.GetController(controller).type = type_t.NewForHint(choices)

        _SignalIssues(self)
        LOGGER.CommitIssues(1)

    def AddUnitSection(self) -> None:
        """"""
        section = free_section_t(
            name=INI_UNIT_SECTION,
            definition="Unit definitions",
            description=f"Units can be used in any other section "
            f"to specify a parameter value as follows: "
            f"numerical_value{INI_UNIT_SEPARATOR}unit, e.g., 1.5'mm.",
            basic=True,
            optional=True,
            category=INI_UNIT_SECTION,
            is_growable=True,
        )
        self.append(section)

    def AddPluginParameter(
        self,
        section: str | any_section_h,
        name: str,
        /,
        *,
        definition: str = "Programmatic plugin parameter",
        description: str = "This parameter is not part of the specification. "
        'It was added programmatically in a "plugin" way.',
        basic: bool = True,
        stripe: any_hint_h | type_t = None,
        default: h.Any = MISSING_REQUIRED_VALUE,
        controlling_value: str = None,
    ) -> None:
        """
        See definition and description above.
        """
        if isinstance(section, str):
            section = self[section]

        parameter = parameter_t(
            name=name,
            definition=definition,
            description=description,
            basic=basic,
            type=stripe,
            default=default,
        )
        self._AddParameter(
            section, parameter, controlling_value, config_t.AddPluginParameter.__name__
        )

    def AddRuntimeParameter(
        self,
        section: str | any_section_h,
        name: str,
        value: str | float,
        /,
        *,
        controlling_value: str = None,
    ) -> parameter_t:
        """
        See definition and description below.
        The existence of such a method is justified by the fact that the parameter
        created can be a "normal" parameter or a unit depending on the section.

        value: float if unit.

        /!\\ Cannot deal with str values with units.
        """
        if isinstance(section, str):
            section = self[section]

        if IsUnitSection(section.name):
            parameter_or_unit_t = unit_t
            basic = True
            converted = value
            stripe = None  # Correctly set by the constructor.
        else:
            parameter_or_unit_t = parameter_t
            basic = section.basic
            converted, _ = ObjectFromStr(value)
            stripe = type(converted)
        definition = "Programmatic runtime parameter"
        description = (
            "This parameter is not part of the specification. "
            "It was added programmatically because it was found in the INI document, "
            "or passed as a command-line argument."
        )
        parameter = parameter_or_unit_t(
            name=name,
            definition=definition,
            description=description,
            basic=basic,
            type=stripe,
            default=converted,  # Just a trick to prevent error if basic is False.
        )
        self._AddParameter(
            section, parameter, controlling_value, config_t.AddRuntimeParameter.__name__
        )

        return parameter

    def _AddParameter(
        self,
        section: any_section_h,
        parameter: parameter_t,
        controlling_value: str,
        caller: str,
        /,
    ) -> None:
        """"""
        if not section.is_growable:
            raise RuntimeError(
                f"{section.name}.{parameter.name}: "
                f"Attempt to add an unspecified parameter to a section accepting none."
            )

        if isinstance(section, free_section_t):
            section.append(parameter)
            return

        controller = section.controller
        if controlling_value is None:
            raise ValueError(
                f"{caller}: A Controlling value must be passed for parameter "
                f"{section.name}.{parameter.name}."
            )

        if controller.primary_value is None:
            # controller_t is immutable, so it must be re-created.
            section.controller = controller_t(
                section=controller.section,
                parameter=controller.parameter,
                primary_value=controlling_value,
            )
            controller = section.controller
            section[controlling_value] = parameters_t((parameter,))
            should_update_controller_choices = True
        elif controlling_value in section:
            section[controlling_value].append(parameter)
            should_update_controller_choices = False
        else:
            section[controlling_value] = parameters_t((parameter,))
            should_update_controller_choices = True

        if should_update_controller_choices:
            nnts = self.GetController(controller).type.annotations
            if nnts.__len__() > 0:
                first_annotation = nnts[0]
            else:
                first_annotation = None
            if (first_annotation is None) or isinstance(first_annotation, choices_t):
                choices = choices_t.NewAnnotatedType(section.controlling_values)
                new_type = type_t.NewForHint(choices)
            else:  # Must be callable_t.
                callable_ = callable_t.NewAnnotatedType(
                    kind=first_annotation.kind,
                    catalog=section.controlling_values,
                    allow_external=first_annotation.allow_external,
                )
                new_type = type_t.NewForHint(callable_)
            self.GetController(controller).type = new_type

    @property
    def section_names(self) -> h.Sequence[str]:
        """"""
        return tuple(_sct.name for _sct in self)

    def GetController(self, controller: controller_t, /) -> parameter_t:
        """"""
        return self[controller.section][controller.parameter]

    def __contains__(self, key: str, /) -> bool:
        """"""
        return any(_elm.name == key for _elm in self)

    def __getitem__(self, key: str, /) -> any_section_h:
        """"""
        for section in self:
            if section.name == key:
                return section

        raise KeyError(f"{key}: Not a section of config.")

    def AsDict(self) -> dict[str, dict[str, h.Any]]:
        """"""
        return {_sct.name: _sct.AsDict() for _sct in self}

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [
            TypeAsRichStr(self),
            f"    [blue]path[/]={self.path}" f"[yellow]:{type(self.path).__name__}[/]",
        ]

        for section in self:
            output.append(text.indent(section.__rich__(), "    "))

        return "\n".join(output)


def _SignalIssues(self: config_t, /) -> None:
    """"""
    with LOGGER.AddedContextLevel(f'Specification "{self.path}"'):
        if self.__len__() == 0:
            LOGGER.StageIssue(f"Empty specification")
            return

        names = self.section_names
        if names.__len__() > set(names).__len__():
            LOGGER.StageIssue("Repeated section names")

        for section in self:
            if isinstance(section, controlled_section_t):
                controller = section.controller
                if controller.section not in self:
                    LOGGER.StageIssue(
                        f"Unspecified parameter "
                        f'"{controller.section}.{controller.parameter}" declared as '
                        f'controller of section "{section.name}"'
                    )
                else:
                    controller_section = self[controller.section]
                    if isinstance(controller_section, controlled_section_t):
                        LOGGER.StageIssue(
                            f'Section "{controller.section}" of parameter '
                            f'"{controller.parameter}", which controls section '
                            f'"{section.name}", is itself controlled'
                        )
                    if controller.parameter not in controller_section:
                        LOGGER.StageIssue(
                            f"Unspecified parameter "
                            f'"{controller.section}.{controller.parameter}" declared '
                            f'as controller of section "{section.name}"'
                        )
                    else:
                        controller_section: free_section_t
                        controller_parameter = controller_section[controller.parameter]
                        if controller_parameter.optional:
                            LOGGER.StageIssue(
                                f"Optional parameter "
                                f'"{controller.section}.{controller.parameter}" '
                                f'declared as controller of section "{section.name}"'
                            )


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import inspect as nspt
import typing as h
from pathlib import Path as pl_path_t

from babelwidget.backend.generic.path_chooser import NewSelectedOutputDocument
from babelwidget.main import backend_t
from babelwidget.main import base_h as library_wgt_h
from conf_ini_g.interface.storage.config import INIConfig, SaveConfigToINIDocument
from conf_ini_g.interface.window.config.action import ActionButtonsLayout
from conf_ini_g.interface.window.config.advanced import AdvancedModeLayout
from conf_ini_g.interface.window.config.title import TitleLayout
from conf_ini_g.interface.window.section.collection import SectionsAndCategories
from conf_ini_g.interface.window.section.main import (
    any_section_h,
    controlled_section_t,
    free_section_t,
)
from conf_ini_g.phase.generic.config import AsStr
from conf_ini_g.phase.specification.config.main import config_t as specification_t
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)
from conf_ini_g.phase.specification.section.specific import (
    any_section_h as any_section_spec_h,
)
from conf_ini_g.phase.specification.section.specific import (
    controlled_section_t as controlled_section_spec_t,
)
from conf_ini_g.phase.specification.section.specific import (
    free_section_t as free_section_spec_t,
)
from conf_ini_g.phase.typed.config.main import (
    TypedFromUntyped,
    WithConsumedUnits,
    typed_config_h,
)
from conf_ini_g.phase.untyped.config import ini_config_h


@d.dataclass(repr=False, eq=False)
class config_t(list[any_section_h]):
    """
    The class cannot use slots because it disables weak referencing, which is required.
    See error message below when using slots:
    TypeError: cannot create weak reference to 'config_t' object
    [...]
    File "[...]conf_ini_g/catalog/interface/window/backend/pyqt5/widget/choices.py", line 41, in SetFunction
        self.clicked.connect(function)
        │                    └ <bound method config_t.ToggleAdvancedMode of <conf_ini_g.interface.window.config.config_t object at [...]>>
        └ <conf_ini_g.catalog.interface.window.backend.pyqt5.widget.choices.radio_choice_wgt_h object at [...]>

    Widget might not cooperate well with list, in which case Python raises the
    following exception: TypeError: multiple bases have instance lay-out conflict
    To be safe, "sections" is a field instead of being part of the class definition.

    _widget: Both an access for interacting with widgets, and a reference keeper to
    prevent autonomous widgets from loosing their "liveness".
    """

    specification: specification_t
    ini_path: pl_path_t | None
    UpdateHistory: h.Callable[[pl_path_t | str], None] | None
    #
    backend: backend_t
    library_wgt: library_wgt_h
    #
    Action: h.Callable[[typed_config_h], None] = None
    #
    _widget: dict[str, library_wgt_h] = d.field(init=False, default_factory=dict)

    @classmethod
    def NewForSpecification(
        cls,
        title: str | None,
        specification: specification_t,
        ini_path: pl_path_t | None,
        backend: backend_t,
        /,
        *,
        history: h.Sequence[str] | None = None,
        UpdateHistory: h.Callable[[pl_path_t | str], None] | None = None,
        action: tuple[str, h.Callable[[typed_config_h], None]] = None,
        advanced_mode: bool = False,
    ) -> h.Self:
        """"""
        if ini_path is not None:
            if UpdateHistory is not None:
                UpdateHistory(ini_path)
            as_str = str(ini_path)
            if history is None:
                history = (as_str,)
            elif as_str not in history:
                history = list(history)
                history.append(as_str)

        if action is None:
            kwargs = {}
        else:
            kwargs = {"Action": action[1]}

        output = cls(
            specification=specification,
            ini_path=ini_path,
            backend=backend,
            library_wgt=backend.base_t(),
            UpdateHistory=UpdateHistory,
            **kwargs,
        )

        # --- Top-level widgets
        (
            title_lyt,
            ini_path_wgt,
            history_button,
            history_menu,
            close_button,
        ) = TitleLayout(
            title,
            specification,
            history,
            backend,
            ini_path,
            output.UpdateWithNewINI,
            output.library_wgt.close,
        )
        advanced_mode_lyt, adv_mode_wgt = AdvancedModeLayout(
            advanced_mode, backend, output.ToggleAdvancedMode
        )
        button_lyt, action_button, action_wgts = ActionButtonsLayout(
            action,
            ini_path is not None,
            backend,
            output.ShowInINIFormat,
            output.SaveConfig,
            output.LaunchAction,
            output.library_wgt.close,
        )
        output._widget["ini_path"] = ini_path_wgt
        output._widget["history_button"] = history_button
        output._widget["history_menu"] = history_menu
        output._widget["adv_mode"] = adv_mode_wgt
        output._widget["action"] = action_button
        output._widget["action_buttons"] = action_wgts
        output._widget["close"] = close_button

        # --- Sections
        sections, controlled_sections, category_selector = SectionsAndCategories(
            specification, None, backend
        )
        output.extend(sections)
        output._AssociateSectionsAndControllers(controlled_sections)
        output._widget["category_selector"] = category_selector

        # --- Layout...
        layout = backend.grid_lyt_t()
        if title_lyt is None:
            first_available_row = 0
        else:
            layout.addLayout(title_lyt, 0, 0, 1, 1)
            first_available_row = 1
        layout.addWidget(category_selector, first_available_row, 0, 1, 1)
        layout.addLayout(advanced_mode_lyt, first_available_row + 1, 0, 1, 1)
        layout.addLayout(button_lyt, first_available_row + 2, 0, 1, 1)

        output.library_wgt.setLayout(layout)
        # --- ...Layout

        return output

    def _AssociateSectionsAndControllers(
        self,
        controlled_sections: h.Sequence[
            tuple[controlled_section_t, controlled_section_spec_t]
        ],
        /,
    ) -> None:
        """"""
        for section, section_spec in controlled_sections:
            controller = section_spec.controller
            # Controller section cannot be controlled, so active parameters are in fact
            # all the parameters.
            parameter = self[controller.section].active_parameters[controller.parameter]
            value_wgt = parameter.value
            if hasattr(value_wgt, "SetFunction"):
                value_wgt.SetFunction(section.page_stack.setCurrentIndex)
            else:
                self.backend.ShowErrorMessage(
                    f"{controller.section}.{controller.parameter}: "
                    f'Controller has no "SetFunction" method; Disabling control.'
                )

    def ReassignCloseButtonTarget(self) -> None:
        """"""
        current = self.library_wgt
        main_window = None
        while current is not None:
            main_window = current
            current = current.parent()

        self._widget["close"].SetFunction(main_window.close)

    def SetValues(
        self, config: typed_config_h, units: dict[tuple[str, str], tuple[str, float]], /
    ) -> None:
        """"""
        for section in self:
            sct_name = section.spec_name
            if isinstance(section, free_section_t):
                for parameter in section:
                    prm_name = parameter.spec_name
                    parameter_spec = self.specification[sct_name][prm_name]
                    typed_section = config[sct_name]
                    if prm_name in typed_section:
                        parameter.value.Assign(
                            typed_section[prm_name], parameter_spec.type
                        )
                        unit = units.get((sct_name, prm_name), None)
                        if unit is not None:
                            parameter.unit.setText(unit[0])
                    elif parameter_spec.optional:
                        parameter.value.Assign(
                            parameter_spec.default, parameter_spec.type
                        )
            else:
                for controlling_value, parameters in section.items():
                    for parameter in parameters:
                        prm_name = parameter.spec_name
                        parameter_spec = self.specification[sct_name][
                            controlling_value
                        ][prm_name]
                        typed_section = config[sct_name]
                        if prm_name in typed_section:
                            parameter.value.Assign(
                                typed_section[prm_name], parameter_spec.type
                            )
                            unit = units.get((sct_name, prm_name), None)
                            if unit is not None:
                                parameter.unit.setText(unit[0])
                        elif parameter_spec.optional:
                            parameter.value.Assign(
                                parameter_spec.default, parameter_spec.type
                            )

    def ToggleAdvancedMode(self, advanced_mode: bool, /) -> None:
        """"""
        for section in self:
            section_spec = self.specification[section.spec_name]
            if section_spec.basic:
                should_check_parameters = True
            elif advanced_mode:
                section.library_wgt.setVisible(True)
                should_check_parameters = True
            else:
                section.library_wgt.setVisible(False)
                should_check_parameters = False

            if should_check_parameters:
                parameters = section.active_parameters
                parameter_specs = self._ActiveParameterSpcOfSection(section_spec)
                for parameter, parameter_spec in zip(parameters, parameter_specs):
                    if not parameter_spec.basic:
                        if advanced_mode:
                            parameter.SetVisible(True)
                        else:
                            parameter.SetVisible(False)

    def UpdateWithNewINI(self, ini_path: pl_path_t | str, /) -> None:
        """"""
        if isinstance(ini_path, str):
            ini_path = pl_path_t(ini_path)

        config_ini = INIConfig(ini_path)
        config_typed, config_units, config_comments, issues = TypedFromUntyped(
            config_ini, None, self.specification
        )
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
            return

        category_selector = self._widget["category_selector"]
        if isinstance(category_selector, self.backend.tabs_t):
            # Note: idx^th layout: category_selector.widget(t_idx).widget().layout().
            while category_selector.count() > 0:
                category_selector.removeTab(0)
        else:
            layout = category_selector.widget().layout()
            while layout.count() > 0:
                layout.itemAt(0).widget().setParent(None)

        self.clear()
        sections, controlled_sections, _ = SectionsAndCategories(
            self.specification, category_selector, self.backend
        )
        self.extend(sections)
        self._AssociateSectionsAndControllers(controlled_sections)
        self.SetValues(config_typed, config_units)
        self.ToggleAdvancedMode(self._widget["adv_mode"].true_btn.isChecked())

        self._widget["ini_path"].Assign(ini_path, None)
        self._widget["history_button"].setEnabled(True)
        if str(ini_path) not in (
            _elm.text() for _elm in self._widget["history_menu"].actions()
        ):
            self._widget["history_menu"].addAction(str(ini_path))

        if self.UpdateHistory is not None:
            self.UpdateHistory(ini_path)

    def _ActiveParameterSpcOfSection(
        self, section_spec: any_section_spec_h, /
    ) -> h.Sequence[parameter_spec_t]:
        """"""
        if isinstance(section_spec, free_section_spec_t):
            output = section_spec
        else:
            controller = section_spec.controller
            controller = self[controller.section].active_parameters[
                controller.parameter
            ]
            output = section_spec.ActiveParameters(controller.Text())

        return output

    def AsINI(self) -> ini_config_h:
        """"""
        return {_elm.spec_name: _elm.AsINI() for _elm in self}

    def LaunchAction(self) -> None:
        """"""
        config_full, units_full, _, issues = TypedFromUntyped(
            self.AsINI(), None, self.specification
        )
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
            return

        typed_config, issues = WithConsumedUnits(config_full, units_full)
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
            return

        self.setEnabled(False)
        self.backend.qt_core_app_t.processEvents()
        try:
            self.Action(typed_config)
        except Exception as exception:
            trace = nspt.trace()[-1]
            context = "\n".join(trace.code_context)
            self.backend.ShowErrorMessage(
                f"{trace.filename}@{trace.lineno}:{trace.function}\n"
                f"{context}\n"
                f"{exception}",
                parent=self.library_wgt,
            )
        self.setEnabled(True)

    def ShowInINIFormat(self) -> None:
        """"""
        config = AsStr(self.AsINI(), color="html")
        self.backend.ShowMessage("INI Config", "<tt>" + config + "<tt/>")

    def SaveConfig(self, new_ini: bool, /) -> None:
        """"""
        if new_ini:
            path = NewSelectedOutputDocument(
                "Save Config As",
                "Save Config As",
                self.backend,
                mode="document",
                valid_types={"Config files": ("ini", "INI")},
            )
        else:
            path = self.ini_path  # Will overwrite current INI document

        if path is not None:
            config = self.AsINI()
            issues = SaveConfigToINIDocument(config, path)
            if issues.__len__() > 0:
                self.backend.ShowErrorMessage("\n".join(issues), parent=self)
            else:
                self.ini_path = path

    def __getattr__(self, attribute: str, /) -> h.Any:
        """
        E.g., used for "show".
        """
        try:
            output = super().__getattribute__(attribute)
        except AttributeError:
            output = getattr(self.library_wgt, attribute)

        return output

    def __contains__(self, key: str, /) -> bool:
        """"""
        return any(_elm.spec_name == key for _elm in self)

    def __getitem__(self, key: str, /) -> any_section_h:
        """"""
        for section in self:
            if section.spec_name == key:
                return section

        raise KeyError(f"{key}: Not a section of config.")


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d

from conf_ini_g.phase.specification.base import base_t
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.section.controller import controller_t
from logger_36 import LOGGER


@d.dataclass(repr=False, eq=False)
class section_t(base_t):
    category: str = "Main"
    optional: bool = False
    is_growable: bool = False  # True by default if no parameters are specified.
    parameters: list[parameter_t] | None = None
    controller: controller_t = None
    alternatives: dict[str, list[parameter_t]] = None

    def __post_init__(self) -> None:
        """"""
        if self.parameters is None:
            self.is_growable = True

        _SignalIssues(self)

    @property
    def controlling_values(self) -> tuple[str, ...]:
        """
        Call only on controlled sections.
        """
        if self.controller.primary_value is None:
            # The controlling values are not mentioned in the specification. They will
            # be set programmatically later on.
            return ()

        if self.alternatives is None:
            # The controlling values are not mentioned in the specification. They are
            # being set programmatically.
            return (self.controller.primary_value,)

        return (self.controller.primary_value,) + tuple(self.alternatives.keys())


def _SignalIssues(self) -> None:
    """"""
    with LOGGER.AddedContextLevel(
        f'Specification of {type(self).__name__} "{self.name}"'
    ):
        if self.parameters is not None:
            valid_name_sets = [tuple(_prm.name for _prm in self.parameters)]
            if self.alternatives is not None:
                for parameters in self.alternatives.values():
                    valid_name_sets.append(tuple(_prm.name for _prm in parameters))
            for valid_name_set in valid_name_sets:
                if valid_name_set.__len__() > set(valid_name_set).__len__():
                    LOGGER.StageIssue(
                        "Repeated parameter names (possibly in alternatives)"
                    )

        basic = self.basic
        optional = self.optional

        if not (basic or optional):
            LOGGER.StageIssue("Section is not basic but not optional")

        if (
            (self.controller is None)
            and (not optional)
            and ((self.parameters is None) or (self.parameters.__len__() == 0))
        ):
            LOGGER.StageIssue("Empty mandatory section")

        if self.parameters is not None:
            n_parameters = 0
            n_basic_prms = 0
            for parameter in self.parameters:
                n_parameters += 1
                if parameter.basic:
                    n_basic_prms += 1

                if parameter.basic and not basic:
                    LOGGER.StageIssue(
                        f'Basic parameter "{parameter.name}" in advanced section'
                    )
                if optional and not parameter.optional:
                    LOGGER.StageIssue(
                        f'Mandatory parameter "{parameter.name}" in optional section'
                    )

            if (n_parameters == 0) and not self.is_growable:
                LOGGER.StageIssue("Empty section not accepting runtime parameters")
            if basic and (n_parameters > 0) and (n_basic_prms == 0):
                LOGGER.StageIssue("Basic section without any basic parameters")

        if self.controller is None:
            if self.alternatives is not None:
                LOGGER.StageIssue("Uncontrolled section with alternatives")
        else:
            if self.controller.section == self.name:
                LOGGER.StageIssue("Section controlled by itself")
            controlling_values = self.controlling_values
            if controlling_values.__len__() > set(controlling_values).__len__():
                LOGGER.StageIssue(
                    "Controlled section with duplicated controlling values"
                )
            if self.parameters is not None:
                for parameter in self.parameters:
                    if not parameter.optional:
                        LOGGER.StageIssue(
                            f'Mandatory parameter "{parameter.name}" in '
                            f"controlled section"
                        )

            if self.controller.primary_value is None:
                if self.alternatives is not None:
                    LOGGER.StageIssue("Controlled section without primary value")
                # else: This will be set programmatically later on.
            elif self.alternatives is None:
                LOGGER.StageIssue("Controlled section without alternatives")


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import types as t
from pathlib import Path as path_t

from babelwidget.main import base_h as library_wgt_h
from conf_ini_g.catalog.boolean import boolean_wgt_t
from conf_ini_g.catalog.callable import callable_wgt_t
from conf_ini_g.catalog.choices import choices_wgt_t
from conf_ini_g.catalog.collection import collection_wgt_t
from conf_ini_g.catalog.none import none_wgt_t
from conf_ini_g.catalog.path import path_wgt_t
from conf_ini_g.catalog.text_line import text_line_t
from str_to_obj.api.catalog import callable_t, choices_t
from str_to_obj.api.type import annotation_t, hint_t, type_t

# Widgets can be mapped from types or annotations. Since annotations are more specific
# than types, they must be placed first to ensure they get a chance to be selected.
_TYPE_WIDGET_TRANSLATOR: dict[type, type[library_wgt_h]] = {
    # Annotations
    callable_t: callable_wgt_t,
    choices_t: choices_wgt_t,
    # Types
    t.NoneType: none_wgt_t,
    t.UnionType: choices_wgt_t,
    bool: boolean_wgt_t,
    float: text_line_t,
    int: text_line_t,
    list: collection_wgt_t,
    path_t: path_wgt_t,
    set: collection_wgt_t,
    str: text_line_t,
    tuple: collection_wgt_t,
}


def ValueWidgetTypeForType(stripe: type_t | hint_t, /) -> type[library_wgt_h]:
    """"""
    base_hint = stripe.type
    nnts = stripe.all_annotations

    for registered_type, widget_type in _TYPE_WIDGET_TRANSLATOR.items():
        if (base_hint is registered_type) or any(
            issubclass(type(_nnt), registered_type) for _nnt in nnts
        ):
            return widget_type

    return text_line_t


def RegisterNewTranslation(
    new_type: type | annotation_t, widget_type: type[library_wgt_h], /
) -> None:
    """"""
    if new_type in _TYPE_WIDGET_TRANSLATOR:
        # Raising an exception is adapted here since it is a developer-oriented function
        raise ValueError(
            f'{new_type.__name__}: Type already registered with "{_TYPE_WIDGET_TRANSLATOR[new_type]}" '
            f"in type-to-widget translations."
        )

    _TYPE_WIDGET_TRANSLATOR[new_type] = widget_type


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import typing as h

from conf_ini_g.interface.storage.parameter import INI_COMMENT_MARKER
from conf_ini_g.interface.storage.section import INI_UNIT_SECTION
from conf_ini_g.phase.specification.config.main import config_t as specification_t
from conf_ini_g.phase.specification.section.specific import (
    controlled_section_t,
    free_section_t,
)
from conf_ini_g.phase.typed.parameter.main import TypedValue
from conf_ini_g.phase.typed.parameter.unit import ConvertedValue
from conf_ini_g.phase.typed.section.units import InitialUnits, UpdateUnits
from conf_ini_g.phase.untyped.config import ini_config_h
from conf_ini_g.phase.untyped.parameter import ValueUnitAndComment
from str_to_obj import INVALID_VALUE

# With interpreted values, and possibly units.
typed_config_h = dict[str, dict[str, h.Any]]


def TypedFromUntyped(
    ini: ini_config_h,
    cmd_line: ini_config_h | None,
    specification: specification_t,
    /,
) -> tuple[
    typed_config_h,
    dict[tuple[str, str], tuple[str, float]],
    dict[tuple[str, str], str],
    list[str],
]:
    """
    Units are not consumed, but returned.
    """
    output = {}
    output_units = {}
    output_comments = {}

    if cmd_line is None:
        configs = (ini,)
    else:
        configs = (ini, cmd_line)
    issues = []
    units = InitialUnits(specification)
    for config in configs:
        for sct_name, parameters in config.items():
            if sct_name not in output:
                output[sct_name] = {}

            if (section_is_units := (sct_name == INI_UNIT_SECTION)) and (
                INI_UNIT_SECTION not in specification
            ):
                specification.AddUnitSection()
            section_spec = specification[sct_name]
            for prm_name, value in parameters.items():
                if section_is_units:
                    value, local_issues = UpdateUnits(units, prm_name, value)
                    if local_issues.__len__() > 0:
                        issues.extend(local_issues)
                        continue
                if isinstance(section_spec, free_section_t) and (
                    prm_name not in section_spec
                ):
                    parameter_spec = specification.AddRuntimeParameter(
                        sct_name, prm_name, value
                    )
                    value = parameter_spec.default
                output[sct_name][prm_name] = value

    # Everything being set, default parameters of controlled sections can be added.
    for config in configs:
        for sct_name, parameters in config.items():
            section_spec = specification[sct_name]
            for prm_name, value in parameters.items():
                if isinstance(section_spec, controlled_section_t):
                    controller = section_spec.controller
                    controlling_value = output[controller.section][controller.parameter]
                    if prm_name not in section_spec[controlling_value]:
                        parameter_spec = specification.AddRuntimeParameter(
                            sct_name,
                            prm_name,
                            value,
                            controlling_value=controlling_value,
                        )
                        output[sct_name][prm_name] = parameter_spec.default

    for section in specification:
        if section.name not in output:
            output[section.name] = {}

        if isinstance(section, free_section_t):
            parameters = section
        else:
            controller = section.controller
            controlling_value = output[controller.section][controller.parameter]
            parameters = section.ActiveParameters(controlling_value)

        output_section = output[section.name]
        for parameter in parameters:
            if parameter.optional and (parameter.name not in output_section):
                output[section.name][parameter.name] = parameter.default

    for sct_name, parameters in output.items():
        if sct_name == INI_UNIT_SECTION:
            continue

        section_spec = specification[sct_name]
        for prm_name, value in parameters.items():
            if isinstance(value, str):
                value, unit, comment = ValueUnitAndComment(value, INI_COMMENT_MARKER)
                if not ((unit is None) or (unit in units)):
                    issues.append(f"/{sct_name}.{prm_name}/ Unknown unit: {unit}.")
                    continue

                if isinstance(section_spec, free_section_t):
                    parameter_spec = section_spec[prm_name]
                else:
                    controller = section_spec.controller
                    controlling_value = output[controller.section][controller.parameter]
                    parameter_spec = section_spec[controlling_value][prm_name]

                value, current_issues = TypedValue(value, parameter_spec.type)
                if current_issues.__len__() > 0:
                    issues.extend(
                        f"/{sct_name}.{prm_name}/ {_iss}" for _iss in current_issues
                    )
                    continue
                output[sct_name][prm_name] = value
                if unit is not None:
                    output_units[(sct_name, prm_name)] = (unit, units[unit])
                if comment is not None:
                    output_comments[(sct_name, prm_name)] = comment

    return output, output_units, output_comments, issues


def WithConsumedUnits(
    config: typed_config_h, units: dict[tuple[str, str], tuple[str, float]], /
) -> tuple[typed_config_h, list[str]]:
    """"""
    output = {_key: dict(_vle) for _key, _vle in config.items()}
    issues = []

    for sct_name, parameters in output.items():
        for prm_name, value in parameters.items():
            unit = units.get((sct_name, prm_name), None)
            if unit is not None:
                converted, unconverted = ConvertedValue(
                    output[sct_name][prm_name], unit[1]
                )
                if unconverted.__len__() > 0:
                    converted = INVALID_VALUE
                    unconverted = ", ".join(unconverted)
                    issues.append(
                        f"{unconverted}: Value(s) do(es) not support unit conversion."
                    )

                output[sct_name][prm_name] = converted

    return output, issues


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import typing as h

from babelwidget.main import backend_t, scroll_container_h
from babelwidget.main import tabs_h as tabs_wgt_h
from conf_ini_g.interface.window.section.main import (
    any_section_h,
    controlled_section_t,
    free_section_t,
)
from conf_ini_g.phase.specification.config.main import config_t as config_spec_t
from conf_ini_g.phase.specification.section.specific import (
    controlled_section_t as controlled_section_spec_t,
)
from conf_ini_g.phase.specification.section.specific import (
    free_section_t as free_section_spec_t,
)


def SectionsAndCategories(
    specification: config_spec_t,
    category_selector: scroll_container_h | tabs_wgt_h | None,
    backend: backend_t,
    /,
) -> tuple[
    list[any_section_h],
    h.Sequence[tuple[controlled_section_t, controlled_section_spec_t]],
    scroll_container_h | tabs_wgt_h,
]:
    """"""
    sections = []
    controlled_sections = []
    categories = {}

    # --- Sections and categories
    for section_spec in specification:
        if isinstance(section_spec, free_section_spec_t):
            section = free_section_t.NewForSpecification(section_spec, backend)
        else:
            section = controlled_section_t.NewForSpecification(
                section_spec,
                section_spec.controller,
                backend,
            )
            if section is not None:
                controlled_sections.append((section, section_spec))
        if section is None:
            continue

        sections.append(section)

        if (category := section_spec.category) not in categories:
            contents = backend.base_t()
            scroll_area = backend.scroll_container_t.NewForWidget(contents)
            layout = backend.vbox_lyt_t()
            contents.setLayout(layout)
            categories[category] = (layout, scroll_area)

        layout = categories[category][0]
        layout.addWidget(section.library_wgt)

    # --- Section dispatch into categories
    missing_category_selector = category_selector is None
    if categories.__len__() > 1:
        if missing_category_selector:
            category_selector = backend.tabs_t()
        for category, (_, scroll_area) in categories.items():
            category_selector.addTab(scroll_area, category)
    elif missing_category_selector:
        category = tuple(categories.keys())[0]
        category_selector = categories[category][1]

    return sections, controlled_sections, category_selector


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import types as t
import typing as h

from babelwidget.main import backend_t
from babelwidget.main import base_h as library_wgt_h
from babelwidget.main import dropdown_choice_h as dropdown_choice_wgt_h
from babelwidget.main import label_h as label_wgt_h
from babelwidget.main import text_line_h as text_line_wgt_h
from conf_ini_g.catalog.choices import choices_wgt_t
from conf_ini_g.catalog.directory import ValueWidgetTypeForType
from conf_ini_g.catalog.multitype import multitype_wgt_t
from conf_ini_g.interface.constant import TYPE_LABEL_WIDTH, TYPE_WIDGET_WIDTH
from conf_ini_g.interface.storage.parameter import INI_UNIT_SEPARATOR
from conf_ini_g.interface.window.generic import FormattedName
from conf_ini_g.interface.window.parameter.type import TypeSelector
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)
from str_to_obj.api.type import hint_t, type_t


@d.dataclass(repr=False, eq=False)
class parameter_t:
    """
    In order to leave the section widget put the name, type, and input widgets of each parameter in columns,
    parameter_t is not a container widget. Instead, it just stores its component widgets for later addition to a layout.
    """

    spec_name: str
    name: label_wgt_h = d.field(init=False, default=None)
    type: label_wgt_h | dropdown_choice_wgt_h = d.field(init=False, default=None)
    value: library_wgt_h = d.field(init=False, default=None)
    unit: text_line_wgt_h = d.field(init=False, default=None)
    comment: str = d.field(init=False, default=None)

    @classmethod
    def NewForSpecification(
        cls,
        parameter_spec: parameter_spec_t,
        backend: backend_t,
        /,
    ) -> h.Self:
        """"""
        output = cls(spec_name=parameter_spec.name)

        formatted_name = FormattedName(parameter_spec.name, " ")
        comment = (
            f"{formatted_name}\n{parameter_spec.definition}.\n\n"
            f"{parameter_spec.description}."
        )
        output.name = backend.label_t(formatted_name)
        output.name.setToolTip(comment)
        output.comment = comment

        stripe = parameter_spec.type
        output.type, output.value = TypeAndValueWidgetsForType(stripe, backend)

        output.unit = backend.text_line_t()

        name_style = "padding-right: 2px;"
        if parameter_spec.optional:
            name_style += "color: gray;"
        output.name.setStyleSheet(name_style)
        output.type.setStyleSheet(name_style)

        return output

    def SetVisible(self, visible: bool, /) -> None:
        """"""
        self.name.setVisible(visible)
        self.type.setVisible(visible)
        self.value.library_wgt.setVisible(visible)
        if self.unit is not None:
            self.unit.setVisible(visible)

    def Text(self) -> str:
        """"""
        text = self.value.Text()
        if self.unit is None:
            return text

        unit = self.unit.Text()
        if unit.__len__() > 0:
            return f"{text}{INI_UNIT_SEPARATOR}{unit}"

        return text


def TypeAndValueWidgetsForType(
    stripe: type_t | hint_t, backend: backend_t, /
) -> tuple[type[library_wgt_h], type[library_wgt_h]]:
    """"""
    if (
        isinstance(stripe, hint_t)
        and (stripe.type is t.UnionType)
        and (stripe.literal_s is None)
    ):
        hint_options = stripe.elements
        type_wgt = TypeSelector(hint_options, backend)
        value_wgt = multitype_wgt_t.NewForHints(
            hint_options,
            type_wgt,
            backend,
        )
    else:
        template = stripe.template_as_str
        if (length := template.__len__()) > TYPE_LABEL_WIDTH:
            shortened = "-" + template[(length - TYPE_LABEL_WIDTH + 1) :]
        else:
            shortened = template
        type_wgt = backend.label_t(shortened)
        type_wgt.setToolTip(template)
        if stripe.literal_s is None:
            widget_type = ValueWidgetTypeForType(stripe)
        else:
            widget_type = choices_wgt_t
        value_wgt = widget_type.NewForSpecification(
            stripe,
            backend,
        )

    type_wgt.setFixedWidth(TYPE_WIDGET_WIDTH)

    return type_wgt, value_wgt


@d.dataclass(init=False, repr=False, eq=False)
class parameters_t(list[parameter_t]):
    def __contains__(self, key: str, /) -> bool:
        """"""
        return any(_elm.spec_name == key for _elm in self)

    def __getitem__(self, key: str, /) -> parameter_t:
        """"""
        for parameter in self:
            if parameter.spec_name == key:
                return parameter

        raise KeyError(f"{key}: Not a valid parameter.")


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import itertools as ittl
import textwrap as text
import typing as h

from conf_ini_g.extension.string import AlignedOnSeparator
from conf_ini_g.phase.specification.base import base_t
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.section.controller import controller_t
from str_to_obj.interface.console import TypeAsRichStr


@d.dataclass(repr=False, eq=False)
class _section_t(base_t):
    """
    Default values only help complying with the constraint: no mandatory attributes
    after the default ones in base_t. In practice, they are mandatory.
    """

    category: str = ""
    optional: bool = False
    is_growable: bool = False

    def __rich_intro__(self) -> list[str]:
        """"""
        return [
            TypeAsRichStr(self),
            *text.indent(super().__rich__(), "    ").splitlines(),
            f"    [blue]Category[/]@=@{self.category}",
            f"    [blue]Optional[/]@=@{self.optional}",
            f"    [blue]Growable[/]@=@{self.is_growable}",
        ]

    @staticmethod
    def __rich_final__(elements: list[str], /) -> str:
        """"""
        elements = AlignedOnSeparator(elements, "@=@", " = ")
        return "\n".join(elements)


@d.dataclass(init=False, repr=False, eq=False)
class parameters_t(list[parameter_t]):
    def __contains__(self, key: str, /) -> bool:
        """"""
        return any(_elm.name == key for _elm in self)

    def __getitem__(self, key: str, /) -> parameter_t:
        """"""
        for parameter in self:
            if parameter.name == key:
                return parameter

        raise KeyError(f"{key}: Not a valid parameter.")


@d.dataclass(slots=True, repr=False, eq=False)
class free_section_t(_section_t, parameters_t):
    parameters: d.InitVar[list[parameter_t] | None] = None

    def __post_init__(self, parameters: list[parameter_t] | None) -> None:
        """"""
        if parameters is not None:
            self.extend(parameters)

    @property
    def all_parameters(self) -> parameters_t:
        """"""
        return self

    def ActiveParameters(self, _: str, /) -> parameters_t:
        """"""
        return self

    def AsDict(self) -> dict[str, h.Any]:
        """"""
        return {_prm.name: _prm.default for _prm in self if _prm.optional}

    def __rich__(self) -> str:
        """"""
        output = self.__rich_intro__()
        output.extend(text.indent(_prm.__rich__(), "    ") for _prm in self)
        return self.__rich_final__(output)


@d.dataclass(slots=True, repr=False, eq=False)
class controlled_section_t(_section_t, dict[str, parameters_t]):
    """
    Default value only help complying with the constraint: no mandatory attributes
    after the default ones in base_t. In practice, it is mandatory.
    """

    controller: controller_t = None
    parameters: d.InitVar[list[parameter_t] | None] = None
    alternatives: d.InitVar[dict[str, list[parameter_t]] | None] = None

    def __post_init__(
        self,
        parameters: list[parameter_t] | None,
        alternatives: dict[str, list[parameter_t]] | None,
    ) -> None:
        """"""
        if parameters is None:
            return

        self[self.controller.primary_value] = parameters_t(parameters)
        for key, value in alternatives.items():
            self[key] = parameters_t(value)

    @property
    def controlling_values(self) -> tuple[str, ...]:
        """"""
        return tuple(self.keys())

    @property
    def all_parameters(self) -> parameters_t:
        """"""
        return parameters_t(ittl.chain(*self.values()))

    def ActiveParameters(self, controlling_value: str, /) -> parameters_t:
        """"""
        if self.__len__() > 0:
            return self[controlling_value]

        # Despite being a controlled section, there can be no alternatives if the
        # section was specified empty. In a GUI context, the section would have
        # been populated programmatically. In a CLI context, it is not.
        return parameters_t()

    def AsDict(self) -> dict[str, h.Any]:
        """"""
        output = {}

        for value, parameters in self.items():
            output.update(
                {
                    (value, _prm.name): _prm.default
                    for _prm in parameters
                    if _prm.optional
                }
            )

        return output

    def __rich__(self) -> str:
        """"""
        output = [
            *self.__rich_intro__(),
            f"    [blue]Controller[/]@=@{self.controller}",
            "    Alternatives:",
        ]

        for ctl_name, parameters in self.items():
            output.append(f"        {ctl_name}")
            for parameter in parameters:
                output.append(text.indent(parameter.__rich__(), "            "))

        return self.__rich_final__(output)


any_section_h = free_section_t | controlled_section_t

"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""
